#include "BigInt.hpp"

namespace big
{
	static const char Bits256[] = {
		0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4,
		5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
		6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
		6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
		8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8
	};

	// returns the degree of the base 2 monic polynomial
	// (the number of bits used to represent the number)
	// eg, 0 0 0 0 1 0 1 1 ... => 28 out of 32 used
	u32 Degree(u32 v)
	{
#if defined(_MSC_VER) && !defined(_DEBUG)
		unsigned long index;
		return _BitScanReverse(&index, v) ? (index + 1) : 0;
#else
		u32 r, t = v >> 16;

		if (t)	r = (r = t >> 8) ? 24 + Bits256[r] : 16 + Bits256[t];
		else 	r = (r = v >> 8) ? 8 + Bits256[r] : Bits256[v];

		return r;
#endif
	}

	// returns the number of limbs that are actually used
	int LimbDegree(const u32 *n, int limbs)
	{
		while (limbs--)
			if (n[limbs])
				return limbs + 1;

		return 0;
	}

	// return bits used
	u32 Degree(const u32 *n, int limbs)
	{
		u32 limb_degree = LimbDegree(n, limbs);
		if (!limb_degree) return 0;
		--limb_degree;

		u32 msl_degree = n[limb_degree];
		
		return msl_degree + limb_degree*32;
	}

	void Set(u32 *lhs, int lhs_limbs, const u32 *rhs, int rhs_limbs)
	{
		int min = lhs_limbs < rhs_limbs ? lhs_limbs : rhs_limbs;

		memcpy(lhs, rhs, min*4);
		memset(&lhs[min], 0, (lhs_limbs - min)*4);
	}
	void Set(u32 *lhs, int limbs, const u32 *rhs)
	{
		memcpy(lhs, rhs, limbs*4);
	}
	void Set32(u32 *lhs, int lhs_limbs, const u32 rhs)
	{
		*lhs = rhs;
		memset(&lhs[1], 0, (lhs_limbs - 1)*4);
	}

#if defined(BIG_ENDIAN)

	// Flip the byte order as needed to make 'n' big-endian for sharing over a network
	void ToLittleEndian(u32 *n, int limbs)
	{
		for (int ii = 0; ii < limbs; ++ii)
		{
			swapLE(n[ii]);
		}
	}

	// Flip the byte order as needed to make big-endian 'n' use the local byte order
	void FromLittleEndian(u32 *n, int limbs)
	{
		// Same operation as ToBigEndian()
		ToLittleEndian(n, limbs);
	}

#endif // BIG_ENDIAN

	bool Less(int limbs, const u32 *lhs, const u32 *rhs)
	{
		for (int ii = limbs-1; ii >= 0; --ii)
		{
			if (lhs[ii] != rhs[ii])
			{
				return lhs[ii] < rhs[ii];
			}
		}

		return false;
	}
	bool Greater(int limbs, const u32 *lhs, const u32 *rhs)
	{
		for (int ii = limbs-1; ii >= 0; --ii)
		{
			if (lhs[ii] != rhs[ii])
			{
				return lhs[ii] > rhs[ii];
			}
		}

		return false;
	}
	bool Equal(int limbs, const u32 *lhs, const u32 *rhs)
	{
		return 0 == memcmp(lhs, rhs, limbs*4);
	}

	bool Less(const u32 *lhs, int lhs_limbs, const u32 *rhs, int rhs_limbs)
	{
		if (lhs_limbs > rhs_limbs)
			do if (lhs[--lhs_limbs] != 0) return false; while (lhs_limbs > rhs_limbs);
		else if (lhs_limbs < rhs_limbs)
			do if (rhs[--rhs_limbs] != 0) return true; while (lhs_limbs < rhs_limbs);
	
		while (lhs_limbs--) if (lhs[lhs_limbs] != rhs[lhs_limbs]) return lhs[lhs_limbs] < rhs[lhs_limbs];
		return false; // equal
	}
	bool Greater(const u32 *lhs, int lhs_limbs, const u32 *rhs, int rhs_limbs)
	{
		if (lhs_limbs > rhs_limbs)
			do if (lhs[--lhs_limbs] != 0) return true; while (lhs_limbs > rhs_limbs);
		else if (lhs_limbs < rhs_limbs)
			do if (rhs[--rhs_limbs] != 0) return false; while (lhs_limbs < rhs_limbs);
	
		while (lhs_limbs--) if (lhs[lhs_limbs] != rhs[lhs_limbs]) return lhs[lhs_limbs] > rhs[lhs_limbs];
		return false; // equal
	}
	bool Equal(const u32 *lhs, int lhs_limbs, const u32 *rhs, int rhs_limbs)
	{
		if (lhs_limbs > rhs_limbs)
			do if (lhs[--lhs_limbs] != 0) return false; while (lhs_limbs > rhs_limbs);
		else if (lhs_limbs < rhs_limbs)
			do if (rhs[--rhs_limbs] != 0) return false; while (lhs_limbs < rhs_limbs);

		while (lhs_limbs--) if (lhs[lhs_limbs] != rhs[lhs_limbs]) return false;
		return true; // equal
	}

	bool Equal32(const u32 *lhs, int lhs_limbs, u32 rhs)
	{
		if (*lhs != rhs) return false;
		while (--lhs_limbs)
			if (*++lhs) return false;
		return true; // equal
	}

	// out = in >>> shift
	// Precondition: 0 <= shift < 31
	void ShiftRight(int limbs, u32 *out, const u32 *in, int shift)
	{
		if (!shift)
		{
			Set(out, limbs, in);
			return;
		}

		u32 carry = 0;

		for (int ii = limbs - 1; ii >= 0; --ii)
		{
			u32 r = in[ii];

			out[ii] = (r >> shift) | carry;

			carry = r << (32 - shift);
		}
	}

	// {out, carry} = in <<< shift
	// Precondition: 0 <= shift < 31
	u32 ShiftLeft(int limbs, u32 *out, const u32 *in, int shift)
	{
		if (!shift)
		{
			Set(out, limbs, in);
			return 0;
		}

		u32 carry = 0;

		for (int ii = 0; ii < limbs; ++ii)
		{
			u32 r = in[ii];

			out[ii] = (r << shift) | carry;

			carry = r >> (32 - shift);
		}

		return carry;
	}

	// lhs += rhs, return carry out
	// precondition: lhs_limbs >= rhs_limbs
	u32 Add(u32 *lhs, int lhs_limbs, const u32 *rhs, int rhs_limbs)
	{
		int ii;
		u64 r = 0;

		for (ii = 0; ii < rhs_limbs; ++ii)
		{
			r >>= 32;
			r += (u64)lhs[ii] + rhs[ii];
			lhs[ii] = (u32)r;
		}

		for (; ii < lhs_limbs && (u32)(r >>= 32) != 0; ++ii)
		{
			r += lhs[ii];
			lhs[ii] = (u32)r;
		}

		return (u32)(r >> 32);
	}

	// out = lhs + rhs, return carry out
	// precondition: lhs_limbs >= rhs_limbs
	u32 Add(u32 *out, const u32 *lhs, int lhs_limbs, const u32 *rhs, int rhs_limbs)
	{
		int ii;
		u64 r = 0;

		for (ii = 0; ii < rhs_limbs; ++ii)
		{
			r >>= 32;
			r += (u64)lhs[ii] + rhs[ii];
			out[ii] = (u32)r;
		}

		for (; ii < lhs_limbs && (u32)(r >>= 32) != 0; ++ii)
		{
			r += lhs[ii];
			out[ii] = (u32)r;
		}

		return (u32)(r >> 32);
	}

	// lhs += rhs, return carry out
	// precondition: lhs_limbs > 0
	u32 Add32(u32 *lhs, int lhs_limbs, u32 rhs)
	{
		u32 n = lhs[0];
		u32 r = n + rhs;
		lhs[0] = r;

		if (r >= n)
			return 0;

		for (int ii = 1; ii < lhs_limbs; ++ii)
			if (++lhs[ii])
				return 0;

		return 1;
	}

	// lhs -= rhs, return borrow out
	// precondition: lhs_limbs >= rhs_limbs
	s32 Subtract(u32 *lhs, int lhs_limbs, const u32 *rhs, int rhs_limbs)
	{
		int ii;
		s64 r = 0;

		for (ii = 0; ii < rhs_limbs; ++ii)
		{
			r >>= 32;
			r += (s64)lhs[ii] - rhs[ii];
			lhs[ii] = (u32)r;
		}

		for (; ii < lhs_limbs && (s32)(r >>= 32) != 0; ++ii)
		{
			r += lhs[ii];
			lhs[ii] = (u32)r;
		}

		return (s32)(r >> 32);
	}

	// out = lhs - rhs, return borrow out
	// precondition: lhs_limbs >= rhs_limbs
	s32 Subtract(u32 *out, const u32 *lhs, int lhs_limbs, const u32 *rhs, int rhs_limbs)
	{
		int ii;
		s64 r = 0;

		for (ii = 0; ii < rhs_limbs; ++ii)
		{
			r >>= 32;
			r += (s64)lhs[ii] - rhs[ii];
			out[ii] = (u32)r;
		}

		for (; ii < lhs_limbs && (s32)(r >>= 32) != 0; ++ii)
		{
			r += lhs[ii];
			out[ii] = (u32)r;
		}

		return (s32)(r >> 32);
	}

	// lhs -= rhs, return borrow out
	// precondition: lhs_limbs > 0, result limbs = lhs_limbs
	s32 Subtract32(u32 *lhs, int lhs_limbs, u32 rhs)
	{
		u32 n = lhs[0];
		u32 r = n - rhs;
		lhs[0] = r;

		if (r <= n)
			return 0;

		for (int ii = 1; ii < lhs_limbs; ++ii)
			if (lhs[ii]--)
				return 0;

		return -1;
	}

	// n = -n
	void Negate(u32 *n, int limbs)
	{
		s64 r = 0;
		for (int ii = 0; ii < limbs; ++ii)
		{
			r >>= 32;
			r -= (s32)n[ii];
			n[ii] = (s32)r;
		}
	}

	// n = ~n, only invert bits up to the MSB, but none above that
	void BitNot(u32 *n, int limbs)
	{
		limbs = LimbDegree(n, limbs);
		if (limbs)
		{
			u32 high = n[--limbs];
			u32 high_degree = 32 - Degree(high);

			n[limbs] = ((u32)(~high << high_degree) >> high_degree);
			while (limbs--) n[limbs] = ~n[limbs];
		}
	}

	// n = ~n, invert all bits, even ones above MSB
	void LimbNot(u32 *n, int limbs)
	{
		while (limbs--) *n++ = ~(*n);
	}

	// lhs ^= rhs
	void Xor(int limbs, u32 *lhs, const u32 *rhs)
	{
		while (limbs--) *lhs++ ^= *rhs++;
	}

	// Return the carry out from A += B << S
    u32 AddLeftShift32(
    	int limbs,		// Number of limbs in parameter A and B
    	u32 *A,			// Large number
    	const u32 *B,	// Large number
    	u32 S)			// 32-bit number
	{
		u64 sum = 0;
		u32 last = 0;

		while (limbs--)
		{
			u32 b = *B++;

			sum = (u64)((b << S) | (last >> (32 - S))) + *A + (u32)(sum >> 32);

			last = b;
			*A++ = (u32)sum;
		}

		return (u32)(sum >> 32) + (last >> (32 - S));
	}

	// Return the carry out from result = A * B
    u32 Multiply32(
    	int limbs,		// Number of limbs in parameter A, result
    	u32 *result,	// Large number
    	const u32 *A,	// Large number
    	u32 B)			// 32-bit number
	{
		u64 p = (u64)A[0] * B;
		result[0] = (u32)p;

		while (--limbs)
		{
			p = (u64)*(++A) * B + (u32)(p >> 32);
			*(++result) = (u32)p;
		}

		return (u32)(p >> 32);
	}

	// Return the carry out from X = X * M + A
    u32 MultiplyAdd32(
    	int limbs,	// Number of limbs in parameter A and B
    	u32 *X,		// Large number
    	u32 M,		// Large number
    	u32 A)		// 32-bit number
	{
		u64 p = (u64)X[0] * M + A;
		X[0] = (u32)p;

		while (--limbs)
		{
			p = (u64)*(++X) * M + (u32)(p >> 32);
			*X = (u32)p;
		}

		return (u32)(p >> 32);
	}

	// Return the carry out from A += B * M
    u32 AddMultiply32(
    	int limbs,		// Number of limbs in parameter A and B
    	u32 *A,			// Large number
    	const u32 *B,	// Large number
    	u32 M)			// 32-bit number
	{
		// This function is roughly 85% of the cost of exponentiation

		// ICC does a better job than my hand-written version by using SIMD instructions,
		// so I use its optimizer instead.
#if !defined(COMPILER_INTEL) && defined(ASSEMBLY_INTEL_SYNTAX)
		ASSEMBLY_BLOCK // VS.NET, x86, 32-bit words
		{
			mov esi, [B]
			mov edi, [A]
			mov eax, [esi]
			mul [M]					; (edx,eax) = [M]*[esi]
			add eax, [edi]			; (edx,eax) += [edi]
			adc edx, 0
			; (edx,eax) = [B]*[M] + [A]

			mov [edi], eax
			; [A] = eax

			mov ecx, [limbs]
			sub ecx, 1
			jz loop_done
loop_head:
				lea esi, [esi + 4]	; ++B
				mov eax, [esi]		; eax = [B]
				mov ebx, edx		; ebx = last carry
				lea edi, [edi + 4]	; ++A
				mul [M]				; (edx,eax) = [M]*[esi]
				add eax, [edi]		; (edx,eax) += [edi]
				adc edx, 0
				add eax, ebx		; (edx,eax) += ebx
				adc edx, 0
				; (edx,eax) = [esi]*[M] + [edi] + (ebx=last carry)

				mov [edi], eax
				; [A] = eax

			sub ecx, 1
			jnz loop_head
loop_done:
			mov [M], edx	; Use [M] to copy the carry into C++ land
		}

		return M;
#else
		// Unrolled first loop
		u64 p = B[0] * (u64)M + A[0];
		A[0] = (u32)p;

		while (--limbs)
		{
			p = (*(++B) * (u64)M + *(++A)) + (u32)(p >> 32);
			A[0] = (u32)p;
		}

		return (u32)(p >> 32);
#endif
	}

	// product = x * y
	void SimpleMultiply(
		int limbs,		// Number of limbs in parameters x, y
		u32 *product,	// Large number; buffer size = limbs*2
		const u32 *x,	// Large number
		const u32 *y)	// Large number
	{
		// Roughly 25% of the cost of exponentiation
		product[limbs] = Multiply32(limbs, product, x, y[0]);

		u32 ctr = limbs;
		while (--ctr)
		{
			++product;
			product[limbs] = AddMultiply32(limbs, product, x, (++y)[0]);
		}
	}

	// product = x ^ 2
	void SimpleSquare(
		int limbs,		// Number of limbs in parameter x
		u32 *product,	// Large number; buffer size = limbs*2
		const u32 *x)	// Large number
	{
		// Seems about 15% faster than SimpleMultiply() in practice
		u32 *cross_product = (u32*)alloca(limbs*2*4);

		// Calculate square-less and repeat-less cross products
		cross_product[limbs] = Multiply32(limbs - 1, cross_product + 1, x + 1, x[0]);
		for (int ii = 1; ii < limbs - 1; ++ii)
		{
			cross_product[limbs + ii] = AddMultiply32(limbs - ii - 1, cross_product + ii*2 + 1, x + ii + 1, x[ii]);
		}

		// Calculate square products
		for (int ii = 0; ii < limbs; ++ii)
		{
			u32 xi = x[ii];
			u64 si = (u64)xi * xi;
			product[ii*2] = (u32)si;
			product[ii*2+1] = (u32)(si >> 32);
		}

		// Multiply the cross product by 2 and add it to the square products
		product[limbs*2 - 1] += AddLeftShift32(limbs*2 - 2, product + 1, cross_product + 1, 1);
	}

	// product = low half of x * y product
	void SimpleMultiplyLowHalf(
		int limbs,		// Number of limbs in parameters x, y
		u32 *product,	// Large number; buffer size = limbs
		const u32 *x,	// Large number
		const u32 *y)	// Large number
	{
		memset(product, 0, limbs*4);

		for (int ii = 0; ii < limbs; ++ii)
		{
			AddMultiply32(limbs - ii, product + ii, x, y[ii]);
		}
	}

	// product = xy
	// memory space for product may not overlap with x,y
    void Multiply(
    	int limbs,		// Number of limbs in x,y
    	u32 *product,	// Product; buffer size = limbs*2
    	const u32 *x,	// Large number; buffer size = limbs
    	const u32 *y)	// Large number; buffer size = limbs
	{
		// Stop recursing under 640 bits or odd limb count
		if (limbs < 30 || (limbs & 1))
		{
			SimpleMultiply(limbs, product, x, y);
			return;
		}

		// Compute high and low products
		Multiply(limbs/2, product, x, y);
		Multiply(limbs/2, product + limbs, x + limbs/2, y + limbs/2);

		// Compute (x1 + x2), xc = carry out
		u32 *xsum = (u32*)alloca((limbs/2)*4);
		u32 xcarry = Add(xsum, x, limbs/2, x + limbs/2, limbs/2);

		// Compute (y1 + y2), yc = carry out
		u32 *ysum = (u32*)alloca((limbs/2)*4);
		u32 ycarry = Add(ysum, y, limbs/2, y + limbs/2, limbs/2);

		// Compute (x1 + x2) * (y1 + y2)
		u32 *cross_product = (u32*)alloca(limbs*4);
		Multiply(limbs/2, cross_product, xsum, ysum);

		// Subtract out the high and low products
		s32 cross_carry = Subtract(cross_product, limbs, product, limbs); 
		cross_carry += Subtract(cross_product, limbs, product + limbs, limbs);

		// Fix the extra high carry bits of the result
		if (ycarry) cross_carry += Add(cross_product + limbs/2, limbs/2, xsum, limbs/2);
		if (xcarry) cross_carry += Add(cross_product + limbs/2, limbs/2, ysum, limbs/2);
		cross_carry += (xcarry & ycarry);

		// Add the cross product into the result
		cross_carry += Add(product + limbs/2, limbs*3/2, cross_product, limbs);

		// Add in the fixed high carry bits
		if (cross_carry) Add32(product + limbs*3/2, limbs/2, cross_carry);
	}

	// product = x^2
	// memory space for product may not overlap with x
    void Square(
    	int limbs,		// Number of limbs in x
    	u32 *product,	// Product; buffer size = limbs*2
    	const u32 *x)	// Large number; buffer size = limbs
	{
		// Stop recursing under 1280 bits or odd limb count
		if (limbs < 40 || (limbs & 1))
		{
			SimpleSquare(limbs, product, x);
			return;
		}

		// Compute high and low squares
		Square(limbs/2, product, x);
		Square(limbs/2, product + limbs, x + limbs/2);

		// Generate the cross product
		u32 *cross_product = (u32*)alloca(limbs*4);
		Multiply(limbs/2, cross_product, x, x + limbs/2);

		// Multiply the cross product by 2 and add it to the result
		u32 cross_carry = AddLeftShift32(limbs, product + limbs/2, cross_product, 1);

		// Roll the carry out up to the highest limb
		if (cross_carry) Add32(product + limbs*3/2, limbs/2, cross_carry);
	}

	// Multiply two large, 2's complement signed numbers: result = a0 * b0
    void SignedMultiply(
    	int limbs,		// Number of limbs in parameters a0,b0
    	u32 *result,	// Output, buffer size = limbs*2
    	const u32 *a0,	// Large number, buffer size = limbs
    	const u32 *b0)	// Large number, buffer size = limbs
	{
		const u32 *a, *b;
		u32 *a1 = (u32*)alloca(limbs*4);
		u32 *b1 = (u32*)alloca(limbs*4);
	
		u32 a_high = a0[limbs-1] & 0x80000000;
	
		if (a_high) {
			Set(a1, limbs, a0, limbs);
			Negate(a1, limbs);
			a = a1;
		} else
			a = a0;
	
		u32 b_high = b0[limbs-1] & 0x80000000;
	
		if (b_high) {
			Set(b1, limbs, b0, limbs);
			Negate(b1, limbs);
			b = b1;
		} else
			b = b0;
	
		Multiply(limbs, result, a, b);
	
		if (a_high ^ b_high)
			Negate(result, limbs);
	}

	// Returns the remainder of N / divisor for a 32-bit divisor
    u32 Modulus32(
    	int limbs,		// Number of limbs in parameter N
    	const u32 *N,	// Large number, buffer size = limbs
    	u32 divisor)	// 32-bit number
	{
		u32 remainder = N[limbs-1] < divisor ? N[limbs-1] : 0;
		u32 counter = N[limbs-1] < divisor ? limbs-1 : limbs;
	
		while (counter--) remainder = (u32)((((u64)remainder << 32) | N[counter]) % divisor);
	
		return remainder;
	}


	/*
	 * 'A' is overwritten with the quotient of the operation
	 * Returns the remainder of 'A' / divisor for a 32-bit divisor
	 * 
	 * Does not check for divide-by-zero
	 */
    u32 Divide32(
    	int limbs,		// Number of limbs in parameter A
    	u32 *A,			// Large number, buffer size = limbs
    	u32 divisor)	// 32-bit number
	{
		u64 r = 0;
		for (int ii = limbs-1; ii >= 0; --ii)
		{
			u64 n = (r << 32) | A[ii];
			A[ii] = (u32)(n / divisor);
			r = n % divisor;
		}
	
		return (u32)r;
	}

	// returns (n ^ -1) Mod 2^32
	u32 MulInverseGF32(u32 n)
	{
		// {u1, g1} = 2^32 / n
		u32 hb = (~(n - 1) >> 31);
		u32 u1 = -(s32)(0xFFFFFFFF / n + hb);
		u32 g1 = ((-(s32)hb) & (0xFFFFFFFF % n + 1)) - n;

		if (!g1) {
			if (n != 1) return 0;
			else return 1;
		}

		u32 q, u = 1, g = n;

		for (;;) {
			q = g / g1;
			g %= g1;

			if (!g) {
				if (g1 != 1) return 0;
				else return u1;
			}

			u -= q*u1;
			q = g1 / g;
			g1 %= g;

			if (!g1) {
				if (g != 1) return 0;
				else return u;
			}

			u1 -= q*u;
		}
	}

	// {q, r} = u / v
	// q is not u or v
	// Return false on divide by zero
	bool Divide(
		const u32 *u,	// numerator, size = u_limbs
		int u_limbs,
		const u32 *v,	// denominator, size = v_limbs
		int v_limbs,
		u32 *q,			// quotient, size = u_limbs
		u32 *r)			// remainder, size = v_limbs
	{
		// calculate v_used and u_used
		int v_used = LimbDegree(v, v_limbs);
		if (!v_used) return false;

		int u_used = LimbDegree(u, u_limbs);

		// if u < v, avoid division
		if (u_used <= v_used && Less(u, u_used, v, v_used))
		{
			// r = u, q = 0
			Set(r, v_limbs, u, u_used);
			Set32(q, u_limbs, 0);
			return true;
		}

		// if v is 32 bits, use faster Divide32 code
		if (v_used == 1)
		{
			// {q, r} = u / v[0]
			Set(q, u_limbs, u);
			Set32(r, v_limbs, Divide32(u_limbs, q, v[0]));
			return true;
		}

		// calculate high zero bits in v's high used limb
		int shift = 32 - Degree(v[v_used - 1]);
		int uu_used = u_used + 1;

		u32 *uu = (u32*)alloca(uu_used*4);
		u32 *vv = (u32*)alloca(v_used*4);

		// shift left to fill high MSB of divisor
		ShiftLeft(v_used, vv, v, shift);
		uu[u_used] = ShiftLeft(u_used, uu, u, shift);

		int q_high_index = uu_used - v_used;

		if (GreaterOrEqual(uu + q_high_index, v_used, vv, v_used))
		{
			Subtract(uu + q_high_index, v_used, vv, v_used);
			Set32(q + q_high_index, u_used - q_high_index, 1);
		}
		else
		{
			Set32(q + q_high_index, u_used - q_high_index, 0);
		}

		u32 *vq_product = (u32*)alloca((v_used+1)*4);

		// for each limb,
		for (int ii = q_high_index - 1; ii >= 0; --ii)
		{
			u32 q_low = (u32)(*(u64*)(uu + ii + v_used - 1) / vv[v_used - 1]);

			vq_product[v_used] = Multiply32(v_used, vq_product, vv, q_low);

			if (Subtract(uu + ii, v_used + 1, vq_product, v_used + 1))
			{
				--q_low;
				if (Add(uu + ii, v_used + 1, vv, v_used) == 0)
				{
					--q_low;
					Add(uu + ii, v_used + 1, vv, v_used);
				}
			}

			q[ii] = q_low;
		}

		memset(r + v_used, 0, (v_limbs - v_used)*4);
		ShiftRight(v_used, r, uu, shift);

		return true;
	}

	// r = u % v
	// Return false on divide by zero
	bool Modulus(
		const u32 *u,	// numerator, size = u_limbs
		int u_limbs,
		const u32 *v,	// denominator, size = v_limbs
		int v_limbs,
		u32 *r)			// remainder, size = v_limbs
	{
		// calculate v_used and u_used
		int v_used = LimbDegree(v, v_limbs);
		if (!v_used) return false;

		int u_used = LimbDegree(u, u_limbs);

		// if u < v, avoid division
		if (u_used <= v_used && Less(u, u_used, v, v_used))
		{
			// r = u, q = 0
			Set(r, v_limbs, u, u_used);
			//Set32(q, u_limbs, 0);
			return true;
		}

		// if v is 32 bits, use faster Divide32 code
		if (v_used == 1)
		{
			// {q, r} = u / v[0]
			//Set(q, u_limbs, u);
			Set32(r, v_limbs, Modulus32(u_limbs, u, v[0]));
			return true;
		}

		// calculate high zero bits in v's high used limb
		int shift = 32 - Degree(v[v_used - 1]);
		int uu_used = u_used + 1;

		u32 *uu = (u32*)alloca(uu_used*4);
		u32 *vv = (u32*)alloca(v_used*4);

		// shift left to fill high MSB of divisor
		ShiftLeft(v_used, vv, v, shift);
		uu[u_used] = ShiftLeft(u_used, uu, u, shift);

		int q_high_index = uu_used - v_used;

		if (GreaterOrEqual(uu + q_high_index, v_used, vv, v_used))
		{
			Subtract(uu + q_high_index, v_used, vv, v_used);
			//Set32(q + q_high_index, u_used - q_high_index, 1);
		}
		else
		{
			//Set32(q + q_high_index, u_used - q_high_index, 0);
		}

		u32 *vq_product = (u32*)alloca((v_used+1)*4);

		// for each limb,
		for (int ii = q_high_index - 1; ii >= 0; --ii)
		{
			u32 q_low = (u32)(*(u64*)(uu + ii + v_used - 1) / vv[v_used - 1]);

			vq_product[v_used] = Multiply32(v_used, vq_product, vv, q_low);

			if (Subtract(uu + ii, v_used + 1, vq_product, v_used + 1))
			{
				//--q_low;
				if (Add(uu + ii, v_used + 1, vv, v_used) == 0)
				{
					//--q_low;
					Add(uu + ii, v_used + 1, vv, v_used);
				}
			}

			//q[ii] = q_low;
		}

		memset(r + v_used, 0, (v_limbs - v_used)*4);
		ShiftRight(v_used, r, uu, shift);

		return true;
	}

	// result = (x * y) (Mod modulus)
	bool MulMod(
		int limbs,			// Number of limbs in x,y,modulus
		const u32 *x,		// Large number x
		const u32 *y,		// Large number y
		const u32 *modulus,	// Large number modulus
		u32 *result)		// Large numbe result
	{
		u32 *product = (u32*)alloca(limbs*2*4);

		Multiply(limbs, product, x, y);

		return Modulus(product, limbs * 2, modulus, limbs, result);
	}

	// Convert bigint to string
	std::string ToStr(const u32 *n, int limbs, int base)
	{
		limbs = LimbDegree(n, limbs);
		if (!limbs) return "0";

		std::string out;
		char ch;

		u32 *m = (u32*)alloca(limbs*4);
		Set(m, limbs, n, limbs);

		while (limbs)
		{
			u32 mod = Divide32(limbs, m, base);
			if (mod <= 9) ch = '0' + mod;
			else ch = 'A' + mod - 10;
			out = ch + out;
			limbs = LimbDegree(m, limbs);
		}

		return out;
	}

	// Convert string to bigint
	// Return 0 if string contains non-digit characters, else number of limbs used
	int ToInt(u32 *lhs, int max_limbs, const char *rhs, u32 base)
	{
		if (max_limbs < 2) return 0;

		lhs[0] = 0;
		int used = 1;

		char ch;
		while ((ch = *rhs++))
		{
			u32 mod;
			if (ch >= '0' && ch <= '9') mod = ch - '0';
			else mod = toupper(ch) - 'A' + 10;
			if (mod >= base) return 0;

			// lhs *= base
			u32 carry = MultiplyAdd32(used, lhs, base, mod);

			// react to running out of room
			if (carry)
			{
				if (used >= max_limbs)
					return 0;

				lhs[used++] = carry;
			}
		}

		if (used < max_limbs)
			Set32(lhs+used, max_limbs-used, 0);

		return used;
	}

	/*
	 * Computes: result = GCD(a, b)  (greatest common divisor)
	 * 
	 * Length of result is the length of the smallest argument
	 */
	void GCD(
		const u32 *a,	//	Large number, buffer size = a_limbs
		int a_limbs,	//	Size of a
		const u32 *b,	//	Large number, buffer size = b_limbs
		int b_limbs,	//	Size of b
		u32 *result)	//	Large number, buffer size = min(a, b)
	{
		int limbs = (a_limbs <= b_limbs) ? a_limbs : b_limbs;

		u32 *g = (u32*)alloca(limbs*4);
		u32 *g1 = (u32*)alloca(limbs*4);

		if (a_limbs <= b_limbs)
		{
			// g = a, g1 = b (mod a)
			Set(g, limbs, a, a_limbs);
			Modulus(b, b_limbs, a, a_limbs, g1);
		}
		else
		{
			// g = b, g1 = a (mod b)
			Set(g, limbs, b, b_limbs);
			Modulus(a, a_limbs, b, b_limbs, g1);
		}

		for (;;) {
			// g = (g mod g1)
			Modulus(g, limbs, g1, limbs, g);

			if (!LimbDegree(g, limbs)) {
				Set(result, limbs, g1, limbs);
				return;
			}

			// g1 = (g1 mod g)
			Modulus(g1, limbs, g, limbs, g1);

			if (!LimbDegree(g1, limbs)) {
				Set(result, limbs, g, limbs);
				return;
			}
		}
	}

	/*
	 * Computes: result = (1/u) (Mod v)
	 * Such that: result * u (Mod v) = 1
	 * Using Extended Euclid's Algorithm (GCDe)
	 * 
	 * This is not always possible, so it will return false iff not possible.
	 */
	bool InvMod(
		const u32 *u,	//	Large number, buffer size = u_limbs
		int u_limbs,	//	Limbs in u
		const u32 *v,	//	Large number, buffer size = limbs
		int limbs,		//	Limbs in modulus(v) and result
		u32 *result)	//	Large number, buffer size = limbs
	{
		u32 *u1 = (u32*)alloca(limbs*4);
		u32 *u3 = (u32*)alloca(limbs*4);
		u32 *v1 = (u32*)alloca(limbs*4);
		u32 *v3 = (u32*)alloca(limbs*4);
		u32 *t1 = (u32*)alloca(limbs*4);
		u32 *t3 = (u32*)alloca(limbs*4);
		u32 *q = (u32*)alloca((limbs + u_limbs)*4);
		u32 *w = (u32*)alloca(limbs*2*4);

		// Unrolled first iteration
		{
			// {q, t3} = u / v
			Divide(u, u_limbs, v, limbs, q, t3);

			// u1 = v1, v1 = t1, u3 = v3, v3 = t3
			Set32(u1, limbs, 0);
			Set32(v1, limbs, 1);
			Set(u3, limbs, v);
			Set(v3, limbs, t3);
		}

		bool iteration = false;

		// loop while v3 != 0
		while (LimbDegree(v3, limbs))
		{
			// {q, t3} = u3 / v3
			Divide(u3, limbs, v3, limbs, q, t3);

			// w = q * v1
			Multiply(limbs, w, q, v1);

			// t1 = u1 + w
			Add(t1, u1, limbs, w, limbs);

			// u1 = v1, v1 = t1, u3 = v3, v3 = t3
			Set(u1, limbs, v1);
			Set(v1, limbs, t1);
			Set(u3, limbs, v3);
			Set(v3, limbs, t3);

			iteration = !iteration;
		}

		if (iteration)
			Set(result, limbs, u1);
		else
			Subtract(result, v, limbs, u1, limbs);

		return Equal32(u3, limbs, 1);
	}

	// Calculates mod_inv from low limb of modulus
	u32 MonModInv(u32 modulus0)
	{
		// mod_inv = -M ^ -1 (Mod 2^32)
		return MulInverseGF32(-(s32)modulus0);
	}

	// Compute n_residue for Montgomery reduction
	void MonInputResidue(
		const u32 *n,		//	Large number, buffer size = n_limbs
		int n_limbs,		//	Number of limbs in n
		const u32 *modulus,	//	Large number, buffer size = m_limbs
		int m_limbs,		//	Number of limbs in modulus
		u32 *n_residue)		//	Result, buffer size = m_limbs
	{
		// p = n * 2^(k*m)
		u32 *p = (u32*)alloca((n_limbs+m_limbs)*4);
		Set(p+m_limbs, n_limbs, n, n_limbs);
		Set32(p, m_limbs, 0);

		// n_residue = p (Mod modulus)
		Modulus(p, n_limbs+m_limbs, modulus, m_limbs, n_residue);
	}

	// result = a * b * r^-1 (Mod modulus) in Montgomery domain
	void MonPro(
		int limbs,				// Number of limbs in each parameter
		const u32 *a_residue,	// Large number, buffer size = limbs
		const u32 *b_residue,	// Large number, buffer size = limbs
		const u32 *modulus,		// Large number, buffer size = limbs
		u32 mod_inv,			// MonModInv() return
		u32 *result)			// Large number, buffer size = limbs
	{
		u32 *t = (u32*)alloca(limbs*2*4);

		Multiply(limbs, t, a_residue, b_residue);
		MonReduce(limbs, t, modulus, mod_inv, result);
	}

	// result = a * r^-1 (Mod modulus) in Montgomery domain
	// The result may be greater than the modulus, but this is okay since
	// the result is still in the RNS.  MonFinish() corrects this at the end.
	void MonReduce(
		int limbs,			// Number of limbs in modulus
		u32 *s,				// Large number, buffer size = limbs*2, gets clobbered
		const u32 *modulus,	// Large number, buffer size = limbs
		u32 mod_inv,		// MonModInv() return
		u32 *result)		// Large number, buffer size = limbs
	{
		// This function is roughly 60% of the cost of exponentiation
		for (int ii = 0; ii < limbs; ++ii)
		{
			u32 q = s[0] * mod_inv;
			s[0] = AddMultiply32(limbs, s, modulus, q);
			++s;
		}

		// Add the saved carries
		if (Add(result, s, limbs, s - limbs, limbs))
		{
			// Reduce the result only when needed
			Subtract(result, limbs, modulus, limbs);
		}
	}

	// result = a * r^-1 (Mod modulus) in Montgomery domain
	void MonFinish(
		int limbs,			// Number of limbs in each parameter
		u32 *n,				// Large number, buffer size = limbs
		const u32 *modulus,	// Large number, buffer size = limbs
		u32 mod_inv)		// MonModInv() return
	{
		u32 *t = (u32*)alloca(limbs*2*4);
		memcpy(t, n, limbs*4);
		memset(t + limbs, 0, limbs*4);

		// Reduce the number
		MonReduce(limbs, t, modulus, mod_inv, n);

		// Fix MonReduce() results greater than the modulus
		if (!Less(limbs, n, modulus))
			Subtract(n, limbs, modulus, limbs);
	}

	// Simple internal version without windowing for small exponents
	static void SimpleMonExpMod(
		const u32 *base,	//	Base for exponentiation, buffer size = mod_limbs
		const u32 *exponent,//	Exponent, buffer size = exponent_limbs
		int exponent_limbs,	//	Number of limbs in exponent
		const u32 *modulus,	//	Modulus, buffer size = mod_limbs
		int mod_limbs,		//	Number of limbs in modulus
		u32 mod_inv,		//	MonModInv() return
		u32 *result)		//	Result, buffer size = mod_limbs
	{
		bool set = false;

		u32 *temp = (u32*)alloca((mod_limbs*2)*4);

		// Run down exponent bits and use the squaring method
		for (int ii = exponent_limbs - 1; ii >= 0; --ii)
		{
			u32 e_i = exponent[ii];

			for (u32 mask = 0x80000000; mask; mask >>= 1)
			{
				if (set)
				{
					// result = result^2
					Square(mod_limbs, temp, result);
					MonReduce(mod_limbs, temp, modulus, mod_inv, result);

					if (e_i & mask)
					{
						// result *= base
						Multiply(mod_limbs, temp, result, base);
						MonReduce(mod_limbs, temp, modulus, mod_inv, result);
					}
				}
				else
				{
					if (e_i & mask)
					{
						// result = base
						Set(result, mod_limbs, base, mod_limbs);
						set = true;
					}
				}
			}
		}
	}

	// Precompute a window for ExpMod() and MonExpMod()
	// Requires 2^window_bits multiplies
	u32 *PrecomputeWindow(const u32 *base, const u32 *modulus, int limbs, u32 mod_inv, int window_bits)
	{
		u32 *temp = (u32*)alloca(limbs*2*4);

		u32 *base_squared = (u32*)alloca(limbs*4);
		Square(limbs, temp, base);
		MonReduce(limbs, temp, modulus, mod_inv, base_squared);

		// precomputed window starts with 000001, 000011, 000101, 000111, ...
		u32 k = (1 << (window_bits - 1));

		u32 *window = new u32[limbs * k];

		u32 *cw = window;
		Set(window, limbs, base);

		while (--k)
		{
			// cw+1 = cw * base^2
			Multiply(limbs, temp, cw, base_squared);
			MonReduce(limbs, temp, modulus, mod_inv, cw + limbs);
			cw += limbs;
		}

		return window;
	};

	// Computes: result = base ^ exponent (Mod modulus)
	// Using Montgomery multiplication with simple squaring method
	// Base parameter must be a Montgomery Residue created with MonInputResidue()
	void MonExpMod(
		const u32 *base,	//	Base for exponentiation, buffer size = mod_limbs
		const u32 *exponent,//	Exponent, buffer size = exponent_limbs
		int exponent_limbs,	//	Number of limbs in exponent
		const u32 *modulus,	//	Modulus, buffer size = mod_limbs
		int mod_limbs,		//	Number of limbs in modulus
		u32 mod_inv,		//	MonModInv() return
		u32 *result)		//	Result, buffer size = mod_limbs
	{
		// Calculate the number of window bits to use (decent approximation..)
		int window_bits = Degree(exponent_limbs);

		// If the window bits are too small, might as well just use left-to-right S&M method
		if (window_bits < 4)
		{
			SimpleMonExpMod(base, exponent, exponent_limbs, modulus, mod_limbs, mod_inv, result);
			return;
		}

		// Precompute a window of the size determined above
		u32 *window = PrecomputeWindow(base, modulus, mod_limbs, mod_inv, window_bits);

		bool seen_bits = false;
		u32 e_bits, trailing_zeroes, used_bits = 0;

		u32 *temp = (u32*)alloca((mod_limbs*2)*4);

		for (int ii = exponent_limbs - 1; ii >= 0; --ii)
		{
			u32 e_i = exponent[ii];

			int wordbits = 32;
			while (wordbits--)
			{
				// If we have been accumulating bits,
				if (used_bits)
				{
					// If this new bit is set,
					if (e_i >> 31)
					{
						e_bits <<= 1;
						e_bits |= 1;

						trailing_zeroes = 0;
					}
					else // the new bit is unset
					{
						e_bits <<= 1;

						++trailing_zeroes;
					}

					++used_bits;

					// If we have used up the window bits,
					if (used_bits == window_bits)
					{
						// Select window index 1011 from "101110"
						u32 window_index = e_bits >> (trailing_zeroes + 1);

						if (seen_bits)
						{
							u32 ctr = used_bits - trailing_zeroes;
							while (ctr--)
							{
								// result = result^2
								Square(mod_limbs, temp, result);
								MonReduce(mod_limbs, temp, modulus, mod_inv, result);
							}

							// result = result * window[index]
							Multiply(mod_limbs, temp, result, &window[window_index * mod_limbs]);
							MonReduce(mod_limbs, temp, modulus, mod_inv, result);
						}
						else
						{
							// result = window[index]
							Set(result, mod_limbs, &window[window_index * mod_limbs]);
							seen_bits = true;
						}

						while (trailing_zeroes--)
						{
							// result = result^2
							Square(mod_limbs, temp, result);
							MonReduce(mod_limbs, temp, modulus, mod_inv, result);
						}

						used_bits = 0;
					}
				}
				else
				{
					// If this new bit is set,
					if (e_i >> 31)
					{
						used_bits = 1;
						e_bits = 1;
						trailing_zeroes = 0;
					}
					else // the new bit is unset
					{
						// If we have processed any bits yet,
						if (seen_bits)
						{
							// result = result^2
							Square(mod_limbs, temp, result);
							MonReduce(mod_limbs, temp, modulus, mod_inv, result);
						}
					}
				}

				e_i <<= 1;
			}
		}

		if (used_bits)
		{
			// Select window index 1011 from "101110"
			u32 window_index = e_bits >> (trailing_zeroes + 1);

			if (seen_bits)
			{
				u32 ctr = used_bits - trailing_zeroes;
				while (ctr--)
				{
					// result = result^2
					Square(mod_limbs, temp, result);
					MonReduce(mod_limbs, temp, modulus, mod_inv, result);
				}

				// result = result * window[index]
				Multiply(mod_limbs, temp, result, &window[window_index * mod_limbs]);
				MonReduce(mod_limbs, temp, modulus, mod_inv, result);
			}
			else
			{
				// result = window[index]
				Set(result, mod_limbs, &window[window_index * mod_limbs]);
				//seen_bits = true;
			}

			while (trailing_zeroes--)
			{
				// result = result^2
				Square(mod_limbs, temp, result);
				MonReduce(mod_limbs, temp, modulus, mod_inv, result);
			}

			//e_bits = 0;
		}

		delete []window;
	}

	// Computes: result = base ^ exponent (Mod modulus)
	// Using Montgomery multiplication with simple squaring method
	void ExpMod(
		const u32 *base,	//	Base for exponentiation, buffer size = base_limbs
		int base_limbs,		//	Number of limbs in base
		const u32 *exponent,//	Exponent, buffer size = exponent_limbs
		int exponent_limbs,	//	Number of limbs in exponent
		const u32 *modulus,	//	Modulus, buffer size = mod_limbs
		int mod_limbs,		//	Number of limbs in modulus
		u32 mod_inv,		//	MonModInv() return
		u32 *result)		//	Result, buffer size = mod_limbs
	{
		u32 *mon_base = (u32*)alloca(mod_limbs*4);
		MonInputResidue(base, base_limbs, modulus, mod_limbs, mon_base);

		MonExpMod(mon_base, exponent, exponent_limbs, modulus, mod_limbs, mod_inv, result);

		MonFinish(mod_limbs, result, modulus, mod_inv);
	}

	// returns b ^ e (Mod m)
	u32 ExpMod(u32 b, u32 e, u32 m)
	{
		// validate arguments
		if (b == 0 || m <= 1) return 0;
		if (e == 0) return 1;

		// find high bit of exponent
		u32 mask = 0x80000000;
		while ((e & mask) == 0) mask >>= 1;

		// seen 1 set bit, so result = base so far
		u32 r = b;

		while (mask >>= 1)
		{
			// VS.NET does a poor job recognizing that the division
			// is just an IDIV with a 32-bit dividend (not 64-bit) :-(

			// r = r^2 (mod m)
			r = (u32)(((u64)r * r) % m);

			// if exponent bit is set, r = r*b (mod m)
			if (e & mask) r = (u32)(((u64)r * b) % m);
		}

		return r;
	}
}
