// version 3

#include "Random.hpp"
#include "BigInt.hpp"
#include "Sha2.hpp"
#include "Clock.hpp"
#include "SecureHandshake.hpp"

#include <iostream>
using namespace std;

using namespace big;


struct NormalContext
{
	u32 N[32];
	u32 mod_inv;

	u32 g[32];

	DigestSHA256 x_hash;

	u32 v[32];

	double timings[32];
};

bool NormalWay(NormalContext &context)
{
	double st;

	st = Clock::usec();
	u32 *x = context.x_hash.getDigest();
	context.timings[0] = Clock::usec() - st;

	st = Clock::usec();
	// client: a = random secret
	u32 a[32];
	Random::ref()->generate(a, sizeof(a));
	context.timings[1] = Clock::usec() - st;

	st = Clock::usec();
	// client: A = g ^ a
	u32 A[32];
	ExpMod(context.g, 32, a, 32, context.N, 32, context.mod_inv, A);
	context.timings[2] = Clock::usec() - st;

	st = Clock::usec();
	// server: b = random secret
	u32 b[32];
	Random::ref()->generate(b, sizeof(a));
	context.timings[3] = Clock::usec() - st;

	st = Clock::usec();
	// server: B = g ^ b
	u32 B[32];
	ExpMod(context.g, 32, b, 32, context.N, 32, context.mod_inv, B);

	// server: B += v
	u32 carry_out = Add(B, 32, context.v, 32);
	if (carry_out || Greater(B, 32, context.N, 32))
	{
		Subtract(B, 32, context.N, 32);
	}
	context.timings[4] = Clock::usec() - st;

	st = Clock::usec();
	// both: u = H(A, B)
	DigestSHA256 u_hash;
	u_hash.performDigest(A, sizeof(A));
	u_hash.performDigest(B, sizeof(B));
	u32 *u = u_hash.endDigest();
	context.timings[5] = Clock::usec() - st;

	st = Clock::usec();
	// client: gb = B - v
	u32 gb[32];
	Set(gb, 32, B, 32);
	u32 borrow_in = Subtract(gb, 32, context.v, 32);
	if (borrow_in)
	{
		Add(gb, 32, context.N, 32);
	}
	context.timings[6] = Clock::usec() - st;

	st = Clock::usec();
	// client: ux = u * x
	u32 ux[16];
	Multiply(8, ux, u, x);
	context.timings[7] = Clock::usec() - st;

	st = Clock::usec();
	// client: aux = a + ux
	u32 aux[32];
	Set(aux, 32, a, 32);
	Add(aux, 32, ux, 16);
	context.timings[8] = Clock::usec() - st;

	st = Clock::usec();
	// client: s = gb ^ aux
	u32 s[32];
	ExpMod(gb, 32, aux, 32, context.N, 32, context.mod_inv, s);
	context.timings[9] = Clock::usec() - st;

	st = Clock::usec();
	// client: k = H{s}
	DigestSHA256 s_hash;
	s_hash.performDigest(s, sizeof(s));
	u32 *k = s_hash.endDigest();
	context.timings[10] = Clock::usec() - st;

	st = Clock::usec();
	// server: vu = v ^ u
	u32 vu[32];
	ExpMod(context.v, 32, u, 8, context.N, 32, context.mod_inv, vu);
	context.timings[11] = Clock::usec() - st;

	st = Clock::usec();
	// server: avu_base = A * vu
	u32 avu[64];
	Multiply(32, avu, A, vu);
	u32 avu_base[32];
	Modulus(avu, 64, context.N, 32, avu_base);
	context.timings[12] = Clock::usec() - st;

	st = Clock::usec();
	// server: s = avu_base ^ b
	u32 s2[32];
	ExpMod(avu_base, 32, b, 32, context.N, 32, context.mod_inv, s2);
	context.timings[13] = Clock::usec() - st;

	st = Clock::usec();
	// server: k2 = H{s2}
	DigestSHA256 s2_hash;
	s2_hash.performDigest(s2, sizeof(s2));
	u32 *k2 = s2_hash.endDigest();
	context.timings[14] = Clock::usec() - st;

	return Equal(s, 32, s2, 32);
}


struct MontgomeryContext
{
	u32 N[32];
	u32 mod_inv;

	u32 g[32];

	DigestSHA256 x_hash;

	u32 v[32];

	double timings[32];
};

bool MontgomeryWay(MontgomeryContext &context)
{
	double st;

	st = Clock::usec();
	u32 *x = context.x_hash.getDigest();
	context.timings[0] = Clock::usec() - st;

	st = Clock::usec();
	// client: a = random secret
	u32 a[32];
	Random::ref()->generate(a, sizeof(a));
	context.timings[1] = Clock::usec() - st;

	st = Clock::usec();
	// client: A = g ^ a
	u32 A[32];
	MonExpMod(context.g, a, 32, context.N, 32, context.mod_inv, A);
	//if (Greater(32, A, context.N)) Subtract(A, 32, context.N, 32);
	context.timings[2] = Clock::usec() - st;

	st = Clock::usec();
	// server: b = random secret
	u32 b[32];
	Random::ref()->generate(b, sizeof(a));
	context.timings[3] = Clock::usec() - st;

	st = Clock::usec();
	// server: B = g ^ b
	u32 B[32];
	MonExpMod(context.g, b, 32, context.N, 32, context.mod_inv, B);
	//if (Greater(32, B, context.N)) Subtract(B, 32, context.N, 32);

	// server: B += v
	Add(B, 32, context.v, 32);
	context.timings[4] = Clock::usec() - st;

	st = Clock::usec();
	// both: u = H(A, B)
	DigestSHA256 u_hash;
	u_hash.performDigest(A, sizeof(A));
	u_hash.performDigest(B, sizeof(B));
	u32 *u = u_hash.endDigest();
	context.timings[5] = Clock::usec() - st;

	st = Clock::usec();
	// client: gb = B - v
	u32 gb[32];
	Set(gb, 32, B, 32);
	Subtract(gb, 32, context.v, 32);
	context.timings[6] = Clock::usec() - st;

	st = Clock::usec();
	// client: ux = u * x
	u32 ux[16];
	Multiply(8, ux, u, x);
	context.timings[7] = Clock::usec() - st;

	st = Clock::usec();
	// client: aux = a + ux
	u32 aux[32];
	Set(aux, 32, a, 32);
	Add(aux, 32, ux, 16);
	context.timings[8] = Clock::usec() - st;

	st = Clock::usec();
	// client: s = gb ^ aux
	u32 s[32];
	MonExpMod(gb, aux, 32, context.N, 32, context.mod_inv, s);
	if (Greater(32, s, context.N)) Subtract(s, 32, context.N, 32);
	context.timings[9] = Clock::usec() - st;

	st = Clock::usec();
	// client: k = H{s}
	DigestSHA256 s_hash;
	s_hash.performDigest(s, sizeof(s));
	u32 *k = s_hash.endDigest();
	context.timings[10] = Clock::usec() - st;

	st = Clock::usec();
	// server: vu = v ^ u
	u32 vu[32];
	MonExpMod(context.v, u, 8, context.N, 32, context.mod_inv, vu);
	//if (Greater(32, vu, context.N)) Subtract(vu, 32, context.N, 32);
	context.timings[11] = Clock::usec() - st;

	st = Clock::usec();
	// server: avu_base = A * vu
	u32 avu_base[32];
	MonPro(32, A, vu, context.N, context.mod_inv, avu_base);
	//if (Greater(32, avu_base, context.N)) Subtract(avu_base, 32, context.N, 32);
	context.timings[12] = Clock::usec() - st;

	st = Clock::usec();
	// server: s2 = avu_base ^ b
	u32 s2[32];
	MonExpMod(avu_base, b, 32, context.N, 32, context.mod_inv, s2);
	if (Greater(32, s2, context.N)) Subtract(s2, 32, context.N, 32);
	context.timings[13] = Clock::usec() - st;

	st = Clock::usec();
	// server: k2 = H{s2}
	DigestSHA256 s2_hash;
	s2_hash.performDigest(s2, sizeof(s2));
	u32 *k2 = s2_hash.endDigest();
	context.timings[14] = Clock::usec() - st;

	return Equal(k, 8, k2, 8);
}

int main()
{
	//Random::ref()->init(1274523365);

	SecureHandshakeServer server;
	SecureHandshakeClient client;

	cout << "Secure Handshake(Offline): Server is generating a key..." << endl;

	// Generate a key on the server side
	if (!server.GenerateKey(2048))
	{
		cout << "Secure Handshake: Unable to generate a key on the server side" << endl;
	}

	cout << "Secure Handshake(Offline): Client is setting the public key..." << endl;

	// Get a copy of the public key
	int public_key_bytes = server.GetPublicKeyBytes();
	char *public_key_buffer = new char[public_key_bytes];
	if (!server.GetPublicKey(public_key_buffer, public_key_bytes))
	{
		cout << "Secure Handshake: Unable to get the public key from the server side" << endl;
	}

	// Give the public key to the client (this is done offline, not during the protocol)
	if (!client.SetPublicKey(public_key_buffer, public_key_bytes))
	{
		cout << "Secure Handshake: Unable to set public key on the client side" << endl;
	}

	delete []public_key_buffer;

	// Gerenate A
	cout << "Secure Handshake(Online): Server is generating the random number A and sending it to the client..." << endl;

	char A[SECURE_HANDSHAKE_A_BYTES];
	server.GenerateA(A, SECURE_HANDSHAKE_A_BYTES);

	// Encrypt the session keys
	cout << "Secure Handshake(Online): Client is generating and encrypting the session keys and sending them to the server..." << endl;

	for (int ii = 0; ii < 10000; ++ii)
	{
		int ESK_bytes = client.GetEncryptedSessionKeyBytes();
		char *EncryptedSessionKey = new char[ESK_bytes];

		char Client_ClientKey[SECURE_HANDSHAKE_KEY_BYTES];
		char Client_ServerKey[SECURE_HANDSHAKE_KEY_BYTES];

		if (!client.EncryptSessionKeys(A, SECURE_HANDSHAKE_A_BYTES, EncryptedSessionKey, ESK_bytes, Client_ServerKey, SECURE_HANDSHAKE_KEY_BYTES, Client_ClientKey, SECURE_HANDSHAKE_KEY_BYTES))
		{
			cout << "Secure Handshake: Unable to encrypt the session keys on the client side" << endl;
		}

		// Decrypt the session keys
		//cout << "Secure Handshake(Online): Server is decrypting and validating the session keys..." << endl;

		char Server_ClientKey[SECURE_HANDSHAKE_KEY_BYTES];
		char Server_ServerKey[SECURE_HANDSHAKE_KEY_BYTES];

		double decrypt_st = Clock::usec();
		if (!server.DecryptSessionKeys(A, SECURE_HANDSHAKE_A_BYTES, EncryptedSessionKey, ESK_bytes, Server_ServerKey, SECURE_HANDSHAKE_KEY_BYTES, Server_ClientKey, SECURE_HANDSHAKE_KEY_BYTES))
		{
			cout << "Secure Handshake: Unable to decrypt the session keys on the server side" << endl;
		}
		else
		{
			double decrypt_et = Clock::usec();
			cout << "Secure Handshake: Server decrypted and validated the key in " << (decrypt_et - decrypt_st) / 1000.0 << " ms" << endl;
		}

		if (0 == memcmp(Server_ClientKey, Client_ClientKey, SECURE_HANDSHAKE_KEY_BYTES))
		{
			//cout << "Secure Handshake: Client key was successfully transferred!" << endl;
		}
		else
		{
			//cout << "Secure Handshake: Client keys on both sides do not match" << endl;
		}

		if (0 == memcmp(Server_ServerKey, Client_ServerKey, SECURE_HANDSHAKE_KEY_BYTES))
		{
			//cout << "Secure Handshake: Server key was successfully transferred!" << endl;
		}
		else
		{
			//cout << "Secure Handshake: Server keys on both sides do not match" << endl;
		}

		delete []EncryptedSessionKey;
	}

	cout << "Secure Handshake: Testing done" << endl;

	//Random::ref()->init(0);

	{
		MontgomeryContext context;

		//generateStrongPseudoPrime(context.N, 32);

		ToInt(context.N, 32,
			  "1086459672876625044081758397343537752233356447767051169745092753115332170877" \
			  "59031702903344116140043399307295740039286923184964118718237760098365968575971485" \
			  "92034688668320514285639800563646734129576299346363617227761833665194149519772117" \
			  "8717925821338845008597132294600926611511882873763202855338520234370327211");
/*
		ToInt(context.N, 32,
			  "9547613142080346291548114488006709727670364183406842071686335016961764526423" \
			  "39214444565956334628238646363878980638787406334566226956173202743777581568370843" \
			  "53659305550000729623663651471494629532599613722290168459031464836621224457551661" \
			  "621445527853489691462182958033139572754692785176812041854358961515809651");
*/
		context.mod_inv = MonModInv(context.N[0]);

		Set32(context.g, 32, 7);
		MonInputResidue(context.g, 32, context.N, 32, context.g);

		const char *name = "name";
		const char *delimiter = ":";
		const char *password = "password";

		context.x_hash.performDigest(name, strlen(name));
		context.x_hash.performDigest(delimiter, strlen(delimiter));
		context.x_hash.performDigest(password, strlen(password));
		u32 *x = context.x_hash.endDigest();

		MonExpMod(context.g, x, 8, context.N, 32, context.mod_inv, context.v);

		double timings_avg[32];
		double timings_min[32];
		OBJCLR(timings_avg);
		OBJCLR(timings_min);

		static const int LOOP_COUNT = 100;

		for (int ii = 0; ii < LOOP_COUNT; ++ii)
		{
			if (!MontgomeryWay(context))
			{
				cout << "Montgomery Way: FAILURE!" << endl;
			}

			for (int ii = 0; ii < 32; ++ii)
			{
				double dt = context.timings[ii];
				timings_avg[ii] += dt;

				if (timings_min[ii] == 0 || dt < timings_min[ii])
					timings_min[ii] = dt;
			}
		}

		for (int ii = 0; ii <= 14; ++ii)
		{
			timings_avg[ii] /= (double)LOOP_COUNT;

			if (timings_avg[ii] > 10)
			{
				cout << "Montgomery Way[" << ii << "]:  min=" << timings_min[ii] / 1000.0 << "ms, avg=" << timings_avg[ii] / 1000.0 << "ms" << endl;
			}
		}
	}

	//Random::ref()->init(0);

	{
		NormalContext context;

		//generateStrongPseudoPrime(context.N, 32);

		ToInt(context.N, 32,
			  "1086459672876625044081758397343537752233356447767051169745092753115332170877" \
			  "59031702903344116140043399307295740039286923184964118718237760098365968575971485" \
			  "92034688668320514285639800563646734129576299346363617227761833665194149519772117" \
			  "8717925821338845008597132294600926611511882873763202855338520234370327211");
/*
		ToInt(context.N, 32,
			  "9547613142080346291548114488006709727670364183406842071686335016961764526423" \
			  "39214444565956334628238646363878980638787406334566226956173202743777581568370843" \
			  "53659305550000729623663651471494629532599613722290168459031464836621224457551661" \
			  "621445527853489691462182958033139572754692785176812041854358961515809651");
*/
		context.mod_inv = MonModInv(context.N[0]);

		Set32(context.g, 32, 7);

		const char *name = "name";
		const char *delimiter = ":";
		const char *password = "password";

		context.x_hash.performDigest(name, strlen(name));
		context.x_hash.performDigest(delimiter, strlen(delimiter));
		context.x_hash.performDigest(password, strlen(password));
		u32 *x = context.x_hash.endDigest();

		ExpMod(context.g, 32, x, 8, context.N, 32, context.mod_inv, context.v);

		double timings_avg[32];
		double timings_min[32];
		OBJCLR(timings_avg);
		OBJCLR(timings_min);

		static const int LOOP_COUNT = 100;

		for (int ii = 0; ii < LOOP_COUNT; ++ii)
		{
			if (!NormalWay(context))
			{
				cout << "Normal Way: FAILURE!" << endl;
			}

			for (int ii = 0; ii < 32; ++ii)
			{
				double dt = context.timings[ii];
				timings_avg[ii] += dt;

				if (timings_min[ii] == 0 || dt < timings_min[ii])
					timings_min[ii] = dt;
			}
		}

		for (int ii = 0; ii <= 14; ++ii)
		{
			timings_avg[ii] /= (double)LOOP_COUNT;

			if (timings_avg[ii] > 10)
			{
				cout << "Normal Way[" << ii << "]:  min=" << timings_min[ii] / 1000.0 << "ms, avg=" << timings_avg[ii] / 1000.0 << "ms" << endl;
			}
		}
	}

	system("pause");

	return 0;
}
